from linebot import LineBotApi
from linebot.models import (
    TemplateSendMessage,
    ButtonsTemplate,
    MessageAction,
    URIAction
)
from eyewitness.config import BBOX
from eyewitness.detection_utils import DetectionResultHandler


class LineAnnotationSender(DetectionResultHandler):
    def __init__(self, audience_ids, channel_access_token, image_url_handler,
                 detection_result_filter=None, detection_method=BBOX):
        """
        Parameters:
        -----------
        audience_ids: set
            line audience ids
        channel_access_token: str
            channel_access_token
        image_url_handler: Callable
            compose drawn image to image_url
        detection_result_filter: Option[Callable]
            a function check if the detection result to sent or not
        detection_method: str
            detection method
        """
        self.audience_ids = audience_ids
        self.line_bot_api = LineBotApi(channel_access_token)
        self.image_url_handler = image_url_handler
        self.detection_result_filter = detection_result_filter
        self._detection_method = detection_method

    @property
    def detection_method(self):
        return self._detection_method

    def _handle(self, detection_result):
        if self.detection_result_filter(detection_result):
            image_url = self.image_url_handler(detection_result.drawn_image_path)
            image_id = str(detection_result.image_id)
            self.send_annotation_button_msg(image_url, image_id)

    def send_annotation_button_msg(self, image_url, image_id):
        """
        sent line botton msg to audience_ids

        Parameters:
        -----------
        img_path: str
            image path to be set
        """
        buttons_msg = TemplateSendMessage(
            alt_text='object detected',
            template=ButtonsTemplate(
                thumbnail_image_url=image_url,
                title='object detected',
                text='help to report result',
                actions=[
                    MessageAction(
                        label='Report Error (錯誤回報)',
                        text=image_id
                    ),
                    URIAction(
                        label='full image (完整圖片)',
                        uri=image_url
                    )
                ]
            )
        )
        self.line_bot_api.multicast(list(self.audience_ids), buttons_msg)



import json
import six
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from eyewitness.config import (FEEDBACK_BBOX, FEEDBACK_FALSE_ALERT)
from eyewitness.image_id import ImageId
from eyewitness.audience_id import AudienceId


BboxObjectFeedback = namedtuple('BboxObjectFeedback',
                                ['x1', 'y1', 'x2', 'y2', 'label', 'meta'])
Type_Serialization_Mapping = {
    FEEDBACK_BBOX: BboxObjectFeedback,
    FEEDBACK_FALSE_ALERT: None,  # pure false alert won't have feedback_msg_objs
}


class FeedbackMsg(object):
    """
    represent the Feedback msg

    init with feedback_dict :
    - audience_id (require): audience information e.g. {platform_id}_{user_id}
    - feedback_method: feedback_method str
    - image_id: image_id obj or string e.g. 'channel1::1541860141::jpg'
    - feedback_meta: feedback_meta str
    - feedback_msg_objs: List of objects List[tuple]
    - receive_time: int

    """
    def __init__(self, feedback_dict):
        # get feedback method
        self.feedback_method = feedback_dict.get('feedback_method', FEEDBACK_FALSE_ALERT)
        self.feedback_type = Type_Serialization_Mapping[self.feedback_method]
        self.feedback_dict = feedback_dict

    @property
    def image_id(self):
        """
        Returns
        -------
        image_id: str
            image_id
        """
        return self.feedback_dict.get('image_id')

    @property
    def audience_id(self):
        """
        Returns
        -------
        audience_id: str
            audience_id
        """
        return self.feedback_dict['audience_id']

    @property
    def receive_time(self):
        """
        Returns
        -------
        receive_time: int
            receive_time
        """
        return self.feedback_dict['receive_time']

    @property
    def feedback_msg_objs(self):
        """
        Returns
        -------
        feedback_msg_objs: List[object]
            List of feedback msg objs
        """
        feedback_msg_objs = self.feedback_dict.get('feedback_msg_objs', [])
        return [self.feedback_type(*i) for i in feedback_msg_objs]

    @property
    def feedback_meta(self):
        """
        Returns
        -------
        feedback_meta: str
            feedback meta information
        """
        return self.feedback_dict.get('feedback_meta')

    @property
    def is_false_alert(self):
        """
        Returns
        -------
        is_false_alert: bool
            image published information is false alert or not
        """
        return self.feedback_dict.get('is_false_alert', True)

    @classmethod
    def from_json(cls, json_str):
        feedback_json_dict = json.loads(json_str)
        feedback_json_dict['audience_id'] = AudienceId.from_str(feedback_json_dict['audience_id'])
        if feedback_json_dict['image_id']:
            feedback_json_dict['image_id'] = ImageId.from_str(feedback_json_dict['image_id'])
        return cls(feedback_json_dict)

    def to_json_dict(self):
        """
        Returns
        -------
        image_dict: dict
            the dict repsentation of detection_result
        """
        json_dict = dict(self.feedback_json_dict)
        json_dict['audience_id'] = str(json_dict['audience_id'])
        if json_dict['image_id']:
            json_dict['image_id'] = str(json_dict['image_id'])
        return json_dict


@six.add_metaclass(ABCMeta)
class FeedbackMsgHandler():
    def handle(self, feedback_msg):
        assert self.feedback_method == feedback_msg.feedback_method
        return self._handle(feedback_msg)

    @abstractmethod
    def _handle(self, feedback_msg):
        """
        abstract method for handle feedback_msg

        Parameters:
        -----------
        feedback_msg: FeedbackMsg
        """
        pass

    @property
    def feedback_method(self):
        raise NotImplementedError



# detection type
BBOX = 'bbox'

# Feedback type
FEEDBACK_BBOX = 'feedback_bbox'
FEEDBACK_FALSE_ALERT = 'feedback_false_alert'


# image producer type
POST_PATH = 'post_path'
POST_BYTES = 'post_bytes'
IN_MEMORY = 'in_memory'



import os
import six
import requests
from abc import ABCMeta

import numpy as np
import pkg_resources

from eyewitness.config import (POST_PATH, POST_BYTES)
from eyewitness.utils import make_path
from eyewitness.image_id import ImageId
from PIL import Image, ImageFont, ImageDraw

DEFAULT_FONT_PATH = pkg_resources.resource_filename('eyewitness', 'font/FiraMono-Medium.otf')


class ImageHandler(object):
    """
    common functions for image processing
    """
    @classmethod
    def save(cls, image, output_path):
        """
        Parameters:
        -----------
        image: PIL.Image
            image obj
        output_path: str
            path to be save
        """
        if isinstance(output_path, ImageId):
            output_path = str(output_path)
        folder = os.path.dirname(output_path)
        make_path(folder)
        image.save(output_path)

    @classmethod
    def read_image_file(cls, image_path):
        """
        Image.open read from file

        Parameters:
        -----------
        image_path: str
            source image path

        Return
        ------
        Image
            PIL.Image instance
        """
        return Image.open(image_path)

    @classmethod
    def read_image_bytes(cls, image_byte):
        """
        Image.open support BytesIO input

        Parameters:
        -----------
        image_path: BytesIO
            read image from ByesIO obj

        Return
        ------
        Image
            PIL.Image instance
        """
        return Image.open(image_byte)

    @classmethod
    def draw_bbox(cls, image, detections, colors=None, font_path=DEFAULT_FONT_PATH):
        """
        draw bbox on to image

        Parameters:
        -----------
        image: Image
            image to be draw
        detections: List[BoundedBoxObject]
            bbox to draw
        colors: Optional[dict]
            color to be used
        font_path: str
            font to be used
        """
        if colors is None:
            colors = {}

        font = ImageFont.truetype(
            font=font_path,
            size=np.floor(3e-2 * image.size[1] + 0.5).astype('int32'))
        thickness = (image.size[0] + image.size[1]) // 300

        for (left, top, right, bottom, predicted_class, score, _) in detections:
            label = '{} {:.2f}'.format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label, font)

            # creating bbox on images
            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, top + 1])

            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i],
                    outline=colors.get(predicted_class, 'red'))
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=colors.get(predicted_class, 'red'))
            draw.text(text_origin, label, fill=(0, 0, 0), font=font)


@six.add_metaclass(ABCMeta)
class ImageProducer():
    def produce_method(self):
        raise NotImplementedError


class PostFilePathImageProducer(ImageProducer):
    def __init__(self, host, protocol='http'):
        self.protocol = protocol
        self.host = host

    def produce_method(self):
        return POST_PATH

    def produce_image(self, image_id, raw_image_path):
        headers = {'image_id': str(image_id),
                   'raw_image_path': raw_image_path}
        requests.post("%s://%s/detect_post_path" % (self.protocol, self.host), headers=headers)


class PostBytesImageProducer(ImageProducer):
    def __init__(self, host, protocol='http'):
        self.protocol = protocol
        self.host = host

    def produce_method(self):
        return POST_BYTES

    def produce_image(self, image_id, image_bytes, raw_image_path=None):
        headers = {'image_id': str(image_id)}

        if raw_image_path:
            headers['raw_image_path'] = raw_image_path

        requests.post(url="%s://%s/detect_post_bytes" % (self.protocol, self.host),
                      headers=headers,
                      data=image_bytes)


def swap_channel_rgb_bgr(image):
    """
    reverse the color channel image
    convert image (w, h, c) with channel rgb -> bgr
    convert image (w, h, c) with channel bgr -> rgb

    Parameters:
    -----------
    image: np.array
    """
    image = image[:, :, ::-1]
    return image



from peewee import (
    CharField,
    TimestampField,
    DoubleField,
    ForeignKeyField,
    IntegerField,
    Model
)

from eyewitness.models.db_proxy import DATABASE_PROXY


class BaseModel(Model):
    class Meta:
        database = DATABASE_PROXY


class ImageInfo(BaseModel):
    image_id = CharField(unique=True, primary_key=True)
    channel = CharField()
    file_format = CharField()
    timestamp = TimestampField()
    raw_image_path = CharField(null=True)
    drawn_image_path = CharField(null=True)


class BboxDetectionResult(BaseModel):
    image_id = ForeignKeyField(ImageInfo)
    x1 = IntegerField()
    x2 = IntegerField()
    y1 = IntegerField()
    y2 = IntegerField()
    label = CharField()
    meta = CharField()
    score = DoubleField()



import json
import six
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from eyewitness.config import BBOX
from eyewitness.image_id import ImageId


BoundedBoxObject = namedtuple(
    'BoundedBoxObject', ['x1', 'y1', 'x2', 'y2', 'label', 'score', 'meta'])
Type_Serialization_Mapping = {BBOX: BoundedBoxObject}


class DetectionResult(object):
    """
    represent detection result of a image.

    init with feedback_dict :
    - detection_method: detection_method str
    - detected_objects: List[tuple]
    - drawn_image_path: path of drawn image str
    - image_id: image_id obj or string e.g. 'channel1::1541860141::jpg'

    """

    def __init__(self, image_dict):
        # get detection method
        self.detection_method = image_dict.get('detection_method', BBOX)
        self.detection_type = Type_Serialization_Mapping[self.detection_method]
        self.image_dict = image_dict

    @property
    def image_id(self):
        """
        Returns
        -------
        image_id: ImageId
            image_id
        """
        return self.image_dict['image_id']

    @property
    def drawn_image_path(self):
        """
        Returns
        -------
        drawn_image_path: str
            drawn_image_path
        """
        return self.image_dict.get('drawn_image_path', '')

    @property
    def detected_objects(self):
        """
        Returns
        -------
        detected_objects: List[object]
            List of detection results
        """
        detected_objects = self.image_dict.get('detected_objects', [])
        return [self.detection_type(*i) for i in detected_objects]

    @classmethod
    def from_json(cls, json_str):
        img_json_dict = json.loads(json_str)
        img_json_dict['image_id'] = ImageId.from_str(img_json_dict['image_id'])
        return cls(img_json_dict)

    def to_json_dict(self):
        """
        Returns
        -------
        image_dict: dict
            the dict repsentation of detection_result
        """
        json_dict = dict(self.image_dict)
        json_dict['image_id'] = str(json_dict['image_id'])
        return json_dict


@six.add_metaclass(ABCMeta)
class DetectionResultHandler():
    @abstractmethod
    def _handle(self, detection_result):
        """
        abstract method for handle DetectionResult

        Parameters:
        -----------
        detection_result: DetectionResult
        """
        pass

    def handle(self, detection_result):
        assert self.detection_method == detection_result.detection_method
        self._handle(detection_result)

    @property
    def detection_method(self):
        raise NotImplementedError



import arrow
import six
from abc import ABCMeta, abstractmethod


class AudienceId(object):
    """
    the target of AudienceId is used to standardize the AudienceId format
    default init method from pattern {platform_id}:{user_id}
    e.g: "line:minhan_hgdfmjg2715"
    """

    def __init__(self, platform_id, user_id):
        """
        Parameters:
        -----------
        platform_id: str
            platform_id of feedback user from
        user_id: str
            id of feedback user
        """
        self.platform_id = platform_id
        self.user_id = user_id

    def __hash__(self):
        return hash(self.platform_id) ^ hash(self.user_id)

    def __str__(self):
        # TODO make a more safer way
        return "{}::{}".format(self.platform_id, self.user_id)

    def __eq__(self, other):
        if isinstance(other, AudienceId):
            return str(self) == str(other)
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, AudienceId):
            return str(self) < str(other)
        return NotImplemented

    @classmethod
    def from_str(cls, audience_id_str):
        # TODO make a more safer way
        platform_id, user_id = audience_id_str.split('::')
        return cls(platform_id=platform_id, user_id=user_id)


@six.add_metaclass(ABCMeta)
class AudienceRegister():
    def register_audience(self, audience_id, meta_dict):
        """
        register audience

        Parameters:
        -----------
        audience_id: AudienceId
            audience information
        meta_dict: dict
            additional information
        """
        register_time = meta_dict.get('register_time', arrow.now().timestamp)
        description = meta_dict.get('description', '')
        self.insert_registered_user(audience_id, register_time, description)

    @abstractmethod
    def insert_registered_user(self, audience_id, register_time, description):
        """
        abstract method for register audience id
        """
        pass



from peewee import Proxy

# Using Peewee proxy to dynamically define database in runtime
# http://docs.peewee-orm.com/en/latest/peewee/database.html#dynamically-defining-a-database
DATABASE_PROXY = Proxy()
